import numpy as np
import os, random, string, re, cctk, ctypes, copy, shutil, time
import subprocess as sp
import multiprocessing as mp

import xtb

from presto import constants, config, constraint

############################
### Directory Parameters ###
############################

# A-Za-z0-9
LETTERS_AND_DIGITS = string.ascii_letters + string.digits

# this is the absolute path to the directory where the run_gaussian.sh script is
GAUSSIAN_SCRIPT_DIRECTORY = config.GAUSSIAN_SCRIPT_DIRECTORY

# this is the absolute path to the directory where the run_xtb.sh script is
XTB_SCRIPT_DIRECTORY = config.XTB_SCRIPT_DIRECTORY

# where the xtb GFNn config files are stored
XTB_PATH = config.XTB_PATH

###########################

class Calculator():
    """
    Generic calculator class.
    """

    def __init__(self):
        pass

    def evaluate(self, atomic_numbers, positions, high_atoms, pipe):
        """
        Return energy, forces
        """
        pass


class NullCalculator(Calculator):
    """
    Does nothing except evaluate constraints. Mostly for testing purposes, but I guess you could also model something insanely boring this way.
    """
    def __init__(self, constraints=list()):
        for c in constraints:
            assert isinstance(c, constraint.Constraint), "{c} is not a valid constraint!"
        self.constraints = constraints

    def evaluate(self, atomic_numbers, positions, high_atoms=None, pipe=None):
        energy = 0
        forces = np.zeros_like(positions).view(cctk.OneIndexedArray)

        for c in self.constraints:
            f, e = c.evaluate(positions)
            energy += e
            forces += f

        if pipe is not None:
            assert isinstance(pipe, mp.connection.Connection), "not a valid Connection instance!"
            pipe.send([energy, forces])
            pipe.close()
        else:
            return energy, forces


class XTBCalculator(Calculator):
    """
    Attributes:
        charge (int):
        multiplicity (int):
        gfn (int or str):
        parallel (int):
    """

    def __init__(self, charge=0, multiplicity=1, gfn=2, parallel=1, constraints=list()):
        assert isinstance(charge, int)
        assert isinstance(multiplicity, int)
        assert isinstance(gfn, (int, str))
        assert isinstance(parallel, int)

        self.charge = charge
        self.multiplicity = multiplicity
        self.gfn = gfn
        self.parallel = parallel

        for c in constraints:
            assert isinstance(c, constraint.Constraint), "{c} is not a valid constraint!"
        self.constraints = constraints

        # select a unique 8-letter string that will
        # be used to identify this calculator
        self.UNIQUE_ID = random.choice(string.ascii_letters)
        for i in range(7):
            self.UNIQUE_ID += random.choice(LETTERS_AND_DIGITS)

        self.COUNTER = 0

    def randomize_id(self):
        """
        Re-randomize the id.
        """
        self.UNIQUE_ID = random.choice(string.ascii_letters)
        for i in range(7):
            self.UNIQUE_ID += random.choice(LETTERS_AND_DIGITS)

    def evaluate(self, atomic_numbers, positions, high_atoms=None, pipe=None, print_timing=False):
        """
        Gets the electronic energy and cartesian forces for the specified geometry.

        Args:
            atomic_numbers (cctk.OneIndexedArray): the atomic numbers (int)
            positions (cctk.OneIndexedArray): the atomic positions in angstroms
            high_atoms (np.ndarray): do nothing with this
            pipe (): for multiprocessing, the connection through which objects should be returned to the parent process
            print_timing (Bool):

        Returns:
            energy (float): in Hartree
            forces (cctk.OneIndexedArray): in amu Å per fs**2
        """
        # set working directory
        old_working_directory = os.getcwd()
        os.chdir(XTB_SCRIPT_DIRECTORY)

        # generate filenames
        attempts = 0
        while True:
            self.COUNTER += 1
            attempts += 1
            this_unique_id = f"{self.UNIQUE_ID}-{self.COUNTER:09d}"
            input_filename = f"presto-{this_unique_id}.xyz"
            if not os.path.isfile(input_filename):
                break
            if attempts > 10000:
                raise ValueError("could not find a unique filename to write!")
        output_filename = f"presto-{this_unique_id}.out"

        # create molecule
        molecule = cctk.Molecule(atomic_numbers, positions, charge=self.charge, multiplicity=self.multiplicity)
        cctk.XYZFile.write_molecule_to_file(input_filename, molecule, title="presto")

        # run job and wait for completion
        command = [
            "bash", "run_xtb.sh",
            f"presto-{this_unique_id}",
            f"{self.charge}",
            f"{self.multiplicity - 1}",
            f"{self.gfn}",
            f"{self.parallel}",
            XTB_PATH,
            f"{self.UNIQUE_ID}.topo",
        ]

        start = time.time()
        process = sp.run(command, capture_output=True)  # redirect stdout and stderr to pipe
        end = time.time()
        if print_timing:
            print(f"\nxtb call {this_unique_id} took {end-start:.3f} s.")

        # get results
        job_directory = f"{XTB_SCRIPT_DIRECTORY}/presto-{this_unique_id}"
        try:
            process.check_returncode()
            os.chdir(job_directory)
            assert os.path.isfile("energy")
            assert os.path.isfile("gradient")
        except Exception as e:
            xtb_input_filename = f"{XTB_SCRIPT_DIRECTORY}/presto-{this_unique_id}/presto-{this_unique_id}.xyz"
            xtb_output_filename = f"{XTB_SCRIPT_DIRECTORY}/presto-{this_unique_id}/presto-{this_unique_id}.out"
            print("========= xtb error ========")
            print("\n===script command issued===")
            print(f"currently in {os.getcwd()}")
            print(command)
            print("\n===xtb input ===")
            print(f">>>{xtb_input_filename}")
            try:
                with open(xtb_input_filename,"r") as f:
                    print(f.read())
            except Exception as e:
                print(e)
            print("\n=== xtb output ===")
            print(f">>>{xtb_output_filename}")
            try:
                with open(xtb_output_filename, "r") as f:
                    print(f.read())
            except Exception as e:
                print(e)
            print("\n===run_xtb.sh script stdout===")
            print(process.stdout.decode("utf-8"))
            print("\n===run_xtb.sh script stderr===")
            print(process.stderr.decode("utf-8"))
            print()
            print("========= xtb error ========")
            raise ValueError(f"command line xtb job {this_unique_id} died with exit code {exit_code} (found_energy_file={found_energy_file}, found_gradient_file={found_gradient_file})!")

        # parse energy
        with open("energy", "r") as f:
            energy_lines = f.read().splitlines()
        energy = energy_lines[1]
        fields = energy.split()
        energy = float(fields[1])

        # parse forces
        with open("gradient", "r") as f:
            gradient_lines = f.read().splitlines()
        forces = []
        for line in gradient_lines:
            fields = line.split()
            if len(fields) == 3:
                x,y,z = float(fields[0]), float(fields[1]), float(fields[2])
                forces.append([-x,-y,-z])
        forces = cctk.OneIndexedArray(forces)
        forces = forces * constants.AMU_A2_FS2_PER_HARTREE_BOHR
        assert len(forces) == molecule.get_n_atoms(), "unexpected number of atoms"

        # delete job directory
        os.chdir(XTB_SCRIPT_DIRECTORY)
        try:
            shutil.rmtree(job_directory)
            pass
        except Exception as e:
            print(e)
            print(f"warning: could not remove ${job_directory} but continuing")

        # restore working directory
        os.chdir(old_working_directory)

        # apply constraints
        for c in self.constraints:
            f, e = c.evaluate(positions)
            energy += e
            forces += f

        # return result
        if pipe is not None:
            assert isinstance(pipe, mp.connection.Connection), "not a valid Connection instance!"
            pipe.send([energy, forces])
            pipe.close()
        else:
            return energy, forces

class GaussianCalculator(Calculator):

    def __init__(
            self,
            charge=0,
            multiplicity=1,
            link0={"mem":"1GB", "nprocshared":"4"},
            route_card="#p hf/3-21g force",
            footer=None,
            constraints=list(),
        ):
        if not re.search("force", route_card):
            raise ValueError("need a force job to calculate forces")

        self.charge = charge
        self.multiplicity = multiplicity
        self.link0 = link0
        self.route_card = route_card
        self.footer = footer

        for c in constraints:
            assert isinstance(c, constraint.Constraint), "{c} is not a valid constraint!"
        self.constraints = constraints

        # select a unique 8-letter string that will
        # be used to identify this calculator
        self.UNIQUE_ID = random.choice(string.ascii_letters)
        for i in range(7):
            self.UNIQUE_ID += random.choice(LETTERS_AND_DIGITS)

        self.COUNTER = 0

    def randomize_id(self):
        self.UNIQUE_ID = random.choice(string.ascii_letters)
        for i in range(7):
            self.UNIQUE_ID += random.choice(LETTERS_AND_DIGITS)

    def evaluate(self, atomic_numbers, positions, high_atoms=None, pipe=None, print_timing=False, qc=False):
        """
        Gets the electronic energy and Cartesian forces for the specified geometry.

        Args:
            atomic_numbers (cctk.OneIndexedArray): the atomic numbers (int)
            positions (cctk.OneIndexedArray): the atomic positions in angstroms
            high_atoms (np.ndarray): do nothing with this
            pipe (): for multiprocessing, the connection through which objects should be returned to the parent process
            print_timing (bool):
            qc (bool): try quadratic convergence for tricky cases, only after default DIIS fails

        Returns:
            energy (float): in Hartree
            forces (cctk.OneIndexedArray): in amu Å per fs**2
        """
        # set working directory
        old_working_directory = os.getcwd()
        os.chdir(GAUSSIAN_SCRIPT_DIRECTORY)

        # generate filenames
        attempts = 0
        while True:
            self.COUNTER += 1
            attempts += 1
            this_unique_id = f"{self.UNIQUE_ID}-{self.COUNTER:09d}"
            input_filename = f"presto-{this_unique_id}.gjf"
            if not os.path.isfile(input_filename):
                break
            if attempts > 10000:
                raise ValueError("could not find a unique filename to write!")
        output_filename = f"presto-{this_unique_id}.out"

        # create molecule
        molecule = cctk.Molecule(atomic_numbers, positions, charge=self.charge, multiplicity=self.multiplicity)

        # write out job
        route_card = self.route_card
        if qc:
            print("warning: initial DIIS convergence failed, trying quadradic convergence")
            route_card = f"{self.route_card} scf=qc"
        cctk.GaussianFile.write_molecule_to_file(input_filename, molecule, route_card, self.link0, self.footer)

        # run job
        start = time.time()
        command = ["bash", "run_gaussian.sh", f"presto-{this_unique_id}"]
        process = sp.Popen(command, stdout=sp.PIPE)
        output, error = process.communicate()
        p_status = process.wait()
        end = time.time()
        output = output.decode("utf-8")
        if p_status != 0:
            raise ValueError(f"gaussian job {this_unique_id} died with exit code {p_status}!")
        if print_timing:
            print(f"\nGaussian call {this_unique_id} took {end-start:.3f} s.")

        energy, forces = None, None
        try:
            # get results
            job_directory = f"{GAUSSIAN_SCRIPT_DIRECTORY}/presto-{this_unique_id}"
            os.chdir(job_directory)
            gaussian_file = cctk.GaussianFile.read_file(output_filename)
            ensemble = gaussian_file.ensemble
            molecule = ensemble.molecules[-1]
            properties_dict = ensemble.get_properties_dict(molecule)
            energy = properties_dict["energy"]
            forces = properties_dict["forces"]
            forces = forces * constants.AMU_A2_FS2_PER_HARTREE_BOHR
        except:
            # worth trying quadratic convergence for those random cases
            try:
                assert qc is False
                self.evaluate(atomic_numbers, positions, high_atoms, pipe, print_timing, qc=True)
            except:
                os.chdir(old_working_directory)
                shutil.copyfile(f"{job_directory}/{output_filename}", output_filename)
                raise ValueError(f"can't parse Gaussian output file - copied to {old_working_directory}/{output_filename} for your perusal")

        # delete job directory
        os.chdir(GAUSSIAN_SCRIPT_DIRECTORY)
        try:
            shutil.rmtree(job_directory)
        except Exception as e:
            print(e)
            print(f"warning: could not remove ${job_directory} but continuing")

        # restore working directory
        os.chdir(old_working_directory)

        # apply constraints
        for c in self.constraints:
            f, e = c.evaluate(positions)
            energy += e
            forces += f

        # return result
        if pipe is not None:
            assert isinstance(pipe, mp.connection.Connection), "not a valid Connection instance!"
            pipe.send([energy, forces])
            pipe.close()
        else:
            return energy, forces

class ONIOMCalculator(Calculator):
    """
    Composes two other calculators, and computes forces according to the ONIOM embedding scheme.
    """
    def __init__(self, high_calculator, low_calculator, constraints=list()):
        assert isinstance(high_calculator, Calculator), "high calculator isn't a proper Calculator!"
        assert isinstance(low_calculator, Calculator), "low calculator isn't a proper Calculator!"
        self.high_calculator = high_calculator
        self.low_calculator = low_calculator

        for c in constraints:
            assert isinstance(c, constraint.Constraint), f"{c} is not a valid constraint!"
        self.high_calculator.constraints = constraints
        self.constraints = constraints

        #### prevent namespace collisions
        self.full_calculator = copy.deepcopy(self.low_calculator)
        self.full_calculator.randomize_id()

        assert self.low_calculator.UNIQUE_ID != self.full_calculator.UNIQUE_ID, "namespace collision!"

    def evaluate(self, atomic_numbers, positions, high_atoms, pipe=None):
        """
        Evaluates the forces according to the ONIOM embedding scheme.
        """
        assert len(high_atoms) > 0, "no point in doing ONIOM without a high layer!"
        assert isinstance(atomic_numbers, cctk.OneIndexedArray), "need to pass one-indexed array for indexing to work properly"
        assert isinstance(positions, cctk.OneIndexedArray), "need to pass one-indexed array for indexing to work properly"

        high_atomic_numbers = atomic_numbers[high_atoms]
        high_positions = positions[high_atoms]

        parent_hh, child_hh = mp.Pipe()
        process_hh = mp.Process(target=self.high_calculator.evaluate, kwargs={
            "atomic_numbers": high_atomic_numbers,
            "positions": high_positions,
            "pipe": child_hh,
        })
        process_hh.start()

        parent_hl, child_hl = mp.Pipe()
        process_hl = mp.Process(target=self.low_calculator.evaluate, kwargs={
            "atomic_numbers": high_atomic_numbers,
            "positions": high_positions,
            "pipe": child_hl,
        })
        process_hl.start()

        parent_ll, child_ll = mp.Pipe()
        process_ll = mp.Process(target=self.full_calculator.evaluate, kwargs={
            "atomic_numbers": atomic_numbers,
            "positions": positions,
            "pipe": child_ll,
        })
        process_ll.start()

        e_hh, f_hh = parent_hh.recv()
        e_hl, f_hl = parent_hl.recv()
        e_ll, f_ll = parent_ll.recv()

        process_hh.join()
        process_hl.join()
        process_ll.join()

        energy = e_hh + e_ll - e_hl
        forces = f_ll
        forces[high_atoms] = f_hh + f_ll[high_atoms] - f_hl

        return energy, forces

class OpenFFCalculator(Calculator):
    def __init__(self, smiles_components, charge=0, multiplicity=1,):
        #### we load these here because it's likely there may be non-OpenMM projects
#        from simtk.unit import unit_definitions as u

        self.charge = charge
        self.multiplicity = multiplicity
        self.smiles_components = smiles_components
        self.simulation = None
        self.mapping = None

#        self.force_units = u.amu * u.angstrom * u.angstrom / u.femtosecond
#        self.energy_units = u.kilocalories_per_mole

    def initialize(self, molecule):
        from simtk import openmm
        import openforcefield as off
        from openforcefield.typing.engines.smirnoff import ForceField
        forcefield = ForceField("openff_unconstrained-1.2.0.offxml")

        #### build OpenForceField topology
        topology = off.topology.Topology()
        self.mapping = list()
        for fragment, smiles in zip(molecule.fragment(), self.smiles_components):
            smiles_mol = cctk.Molecule.new_from_smiles(smiles).assign_connectivity()
            offm = off.topology.Molecule.from_smiles(smiles)
            topology.add_molecule(offm)

            fragment.assign_connectivity()
            iso, mapping = cctk.Molecule.are_isomorphic(fragment, smiles_mol, return_ordering=True)
            assert iso, f"fragment is not isomorphic to smiles {smiles}"
            self.mapping += mapping

        print([[a.topology_atom_index for a in b.atoms] for b in topology.topology_bonds])
        print(mapping)

        #### set up OpenMM system
        forcefield.get_parameter_handler('Electrostatics').method = 'PME'
        system = forcefield.create_openmm_system(topology)
        integrator = openmm.VerletIntegrator(1.0) # we won't use this
        simulation = openmm.app.Simulation(topology.to_openmm(), system, integrator)
        self.simulation = simulation

    def evaluate(self, atomic_numbers, positions, high_atoms=None, pipe=None, print_timing=False):
        """
        Gets the electronic energy and Cartesian forces for the specified geometry.

        Args:
            atomic_numbers (cctk.OneIndexedArray): the atomic numbers (int)
            positions (cctk.OneIndexedArray): the atomic positions in angstroms
            high_atoms (np.ndarray): do nothing with this
            pipe (): for multiprocessing, the connection through which objects should be returned to the parent process
            print_timing (Bool):

        Returns:
            energy (float): in Hartree
            forces (cctk.OneIndexedArray): in amu Å per fs**2
        """

        if self.simulation is None:
            molecule = cctk.Molecule(atomic_numbers, positions).assign_connectivity()
            self.initialize(molecule)

        positions = positions[self.mapping] * 0.1 # convert to nm
        positions = positions.view(np.ndarray).tolist()

        # run job
        start = time.time()

        s = self.simulation
        s.context.setPositions(positions)
        state = s.context.getState(getEnergy=True, getForces=True)
        energy = state.getPotentialEnergy() #/ self.energy_units * constants.HARTREE_PER_KCAL
        forces = state.getForces(asNumpy=True) #/ self.force_units

        end = time.time()
        if print_timing:
            print(f"\nOpenMM call took {end-start:.3f} s.")

        # apply constraints
        for c in self.constraints:
            f, e = c.evaluate(positions)
            energy += e
            forces += f

        # return result
        if pipe is not None:
            assert isinstance(pipe, mp.connection.Connection), "not a valid Connection instance!"
            pipe.send([energy, forces])
            pipe.close()
        else:
            return energy, forces





import numpy as np
import math, copy, cctk, os, re, logging, time

import h5py
import presto

logger = logging.getLogger(__name__)

class Trajectory():
    """

    Attributes:
        timestep (float): in fs
        frames (list of presto.Frame):
        stop_time (float): how long to run for

        high_atoms (np.ndarray): to calculate at high level of theory, list of 1-indexed atom numbers
        active_atoms (np.ndarray): non-frozen atoms, list of 1-indexed atom numbers

        atomic_numbers (cctk.OneIndexedArray): list of atomic numbers
        masses (cctk.OneIndexedArray): list of masses

        calculator (presto.Calculator):
        integrator (presto.Integrator):

        finished (bool):
        forwards (bool):

        checkpoint_filename (str):
    """

    def __init__(self, calculator=None, integrator=None, timestep=None, atomic_numbers=None, high_atoms=None, forwards=True, checkpoint_filename=None, stop_time=None, **kwargs):
        if checkpoint_filename is not None:
            assert isinstance(checkpoint_filename, str), "need string for file"
        self.checkpoint_filename = checkpoint_filename
        if self.has_checkpoint():
            if "load_frames" in kwargs:
                self.load_from_checkpoint(kwargs["load_frames"])
            else:
                self.load_from_checkpoint()

        if calculator is not None:
            assert isinstance(calculator, presto.calculators.Calculator), "need a valid calculator!"
        if integrator is not None:
            assert isinstance(integrator, presto.integrators.Integrator), "need a valid integrator!"

        self.calculator = calculator
        self.integrator = integrator

        if atomic_numbers is not None:
            assert isinstance(atomic_numbers, cctk.OneIndexedArray), "atomic numbers must be cctk 1-indexed array!"
            self.atomic_numbers = atomic_numbers
        elif not hasattr(self, "atomic_numbers"):
            raise ValueError("no atomic numbers specified")

        if not hasattr(self, "finished"):
            self.finished = False

        if high_atoms is not None:
            assert isinstance(high_atoms, np.ndarray), "high_atoms must be np.ndarray!"
            self.high_atoms = high_atoms
        else:
            self.high_atoms = None

        active_atoms = None
        if "active_atoms" in kwargs:
            active_atoms = kwargs["active_atoms"]
            assert isinstance(active_atoms, np.ndarray), "active_atoms must be np.ndarray!"
            self.active_atoms = active_atoms
        elif "inactive_atoms" in kwargs:
            self.set_inactive_atoms(kwargs["inactive_atoms"])
        else:
            # assume all atoms are active
            self.set_inactive_atoms(np.ndarray([]))

        if timestep is not None:
            assert timestep > 0, "can't have timestep ≤ 0!"
            self.timestep = float(timestep)

        if not hasattr(self, "masses"):
            self.masses = cctk.OneIndexedArray([float(cctk.helper_functions.draw_isotopologue(z)) for z in atomic_numbers])

        if not hasattr(self, "frames"):
            self.frames = []

        if forwards is not None:
            assert isinstance(forwards, bool), "forwards must be bool"
            self.forwards = forwards
        elif not hasattr(self, "forwards"):
            self.forwards = True

        if not hasattr(self, "stop_time"):
            assert (isinstance(stop_time, float)) or (isinstance(stop_time, int)), "stop_time needs to be numeric!"
            assert stop_time > 0, "stop_time needs to be positive!"
            self.stop_time = stop_time

    def __str__(self):
        return f"Trajectory({len(self.frames)} frames)"

    def __repr__(self):
        return f"Trajectory({len(self.frames)} frames)"

    def set_inactive_atoms(self, inactive_atoms):
        """
        Since sometimes it's easier to specify the inactive atoms than the inactive atoms, this method updates ``self.active_atoms`` with the complement of ``inactive_atoms``.
        """
        assert isinstance(inactive_atoms, (list, np.ndarray)), "Need list of atoms!"
        active_atoms = list(range(1, len(self.atomic_numbers)+1))
        if len(inactive_atoms):
            for atom in inactive_atoms:
                active_atoms.remove(atom)
        active_atoms = np.array(active_atoms)
        self.active_atoms = active_atoms

    def run(self, checkpoint_interval=10, keep_all=False, time=None, **kwargs):
        """
        Run the trajectory.

        Args:
            checkpoint_interval (int): interval at which to save (in frames, not fs)
            keep_all (bool): whether or not to keep all frames in memory
            num (float): total time to run for -- default is None, implying trajectory should be run until finished
        """
        assert isinstance(self.calculator, presto.calculators.Calculator), "need a valid calculator!"
        assert isinstance(self.integrator, presto.integrators.Integrator), "need a valid integrator!"
        if self.checkpoint_filename is None:
            if "checkpoint_filename" in kwargs:
                self.checkpoint_filename = kwargs["checkpoint_filename"]
            else:
                raise ValueError("no checkpoint filename given!")

        if self.has_checkpoint():
            self.load_from_checkpoint()

        if len(self.frames) == 0:
            self.initialize(**kwargs)
        else:
            self.frames = [self.frames[-1]] # we don't load all frames while running to keep things fast

        if self.finished:
            logger.info("Trajectory already finished!")
            return self
        else:
            logger.info("Propagating trajectory.")
            self.propagate(checkpoint_interval, keep_all=keep_all, time=time)
            self.save(keep_all=keep_all)
            return self

    def initialize(self):
        """
        Adds first frame with randomly-initialized velocities. Should call ``self.save()``.
        """
        pass

    def propagate(self, checkpoint_interval, keep_all=False, time=None):
        """
        Runs trajectories, checking for completion and saving as necessary.

        Will call ``self.save()`` every ``checkpoint_interval`` frames.

        Args:
            checkpoint_interval (int): interval at which to save (in frames, not fs)
            keep_all (bool): whether or not to keep all frames in memory
            time (float): total time to run for -- default is None, implying trajectory should be run until finished
        """
        pass

    def has_checkpoint(self, max_wait=100):
        if self.checkpoint_filename is None:
            return False
        if os.path.exists(self.checkpoint_filename):
            return True
        else:
            return False

        #### TODO - prevent lockfile collisions here and in num_frames

#            lockfile = f"{self.checkpoint_filename}.lock"
#
#            while os.path.exists(lockfile):
#                max_wait += -1
#                time.sleep(1)
#
#                if max_wait < 0:
#                    raise ValueError(f"can't get to file {self.checkpoint_filename}; max_wait exceeded!")
#
#            # create new lockfile
#            with open(lockfile, "w") as lf:
#                pass
#
#            status = False
#            with h5py.File(self.checkpoint_filename, "r+") as h5:
#                all_energies = h5.get("all_energies")
#                if all_energies is None:
#                    status = False
#                elif len(all_energies) > 0:
#                    status = True
#                else:
#                    status = False
#
#            os.remove(lockfile)
#            return status
#
#        else:
#            return False

    def load_from_checkpoint(self, frames=slice(None), max_wait=100):
        """
        Loads frames from ``self.checkpoint_filename``.

        Args:
            frames (Slice object): if not all frames are desired, a Slice object can be passed

        Returns:
            nothing
        """
        assert self.has_checkpoint(), "can't load without checkpoint file"

        lockfile = f"{self.checkpoint_filename}.lock"
        while os.path.exists(lockfile):
            max_wait += -1
            time.sleep(1)

            if max_wait < 0:
                raise ValueError(f"can't get to file {self.checkpoint_filename}; max_wait exceeded!")

        # create new lockfile
        with open(lockfile, "w") as lf:
            pass

        with h5py.File(self.checkpoint_filename, "r") as h5:
            atomic_numbers = h5.attrs["atomic_numbers"]
            self.atomic_numbers = cctk.OneIndexedArray(atomic_numbers)

            masses = h5.attrs["masses"]
            self.masses = cctk.OneIndexedArray(masses)

            self.finished = h5.attrs['finished']
            self.forwards = h5.attrs['forwards']

            all_energies = h5.get("all_energies")[frames]
            all_positions = h5.get("all_positions")[frames]
            all_velocities= h5.get("all_velocities")[frames]
            all_accels = h5.get("all_accelerations")[frames]
            temperatures = h5.get("bath_temperatures")[frames]

            assert len(all_positions) == len(all_energies)
            assert len(all_velocities) == len(all_energies)
            assert len(all_accels) == len(all_energies)

            self.frames = []
            for i, e in enumerate(all_energies):
                self.frames.append(presto.frame.Frame(
                    self,
                    all_positions[i].view(cctk.OneIndexedArray),
                    all_velocities[i].view(cctk.OneIndexedArray),
                    all_accels[i].view(cctk.OneIndexedArray),
                    energy=e,
                    bath_temperature=temperatures[i]
                ))
        logger.info(f"Loaded trajectory from checkpoint file {self.checkpoint_filename} -- {len(self.frames)} frames read.")

        os.remove(lockfile)
        return

    def num_frames(self):
        if self.has_checkpoint():
            num = 0
            with h5py.File(self.checkpoint_filename, "r") as h5:
                num = len(h5.get("all_energies"))
            return num
        else:
            return len(self.frames)

    def save(self, keep_all=False, max_wait=100):
        if self.checkpoint_filename is None:
            raise ValueError("can't save without checkpoint filename")

        lockfile = f"{self.checkpoint_filename}.lock"
        while os.path.exists(lockfile):
            max_wait += -1
            time.sleep(1)

            if max_wait < 0:
                raise ValueError(f"can't get to file {self.checkpoint_filename}; max_wait exceeded!")

        # create new lockfile
        with open(lockfile, "w") as lf:
            pass

        if self.has_checkpoint():
            with h5py.File(self.checkpoint_filename, "r+") as h5:
                n_atoms = len(self.atomic_numbers)
                h5.attrs['finished'] = self.finished
                all_energies = h5.get("all_energies")

                old_n_frames = len(all_energies)
                new_n_frames = len(self.frames) - 1
                now_n_frames = new_n_frames + old_n_frames

                if new_n_frames == 0:
                    os.remove(lockfile)
                    return
                assert new_n_frames > 0, f"we can't write negative frames ({old_n_frames} previously in {self.checkpoint_filename}, but now only {now_n_frames})"

                new_energies = np.asarray([frame.energy for frame in self.frames[-new_n_frames-1:]])
                all_energies.resize((now_n_frames,))
                all_energies[-new_n_frames-1:] = new_energies

                new_positions = np.stack([frame.positions for frame in self.frames[-new_n_frames:]])
                all_positions = h5.get("all_positions")
                all_positions.resize((now_n_frames,n_atoms,3))
                all_positions[-new_n_frames:] = new_positions

                new_velocities= np.stack([frame.velocities for frame in self.frames[-new_n_frames:]])
                all_velocities = h5.get("all_velocities")
                all_velocities.resize((now_n_frames,n_atoms,3))
                all_velocities[-new_n_frames:] = new_velocities

                new_accels = np.stack([frame.accelerations for frame in self.frames[-new_n_frames:]])
                all_accels = h5.get("all_accelerations")
                all_accels.resize((now_n_frames,n_atoms,3))
                all_accels[-new_n_frames:] = new_accels

                new_temps = np.stack([frame.bath_temperature for frame in self.frames[-new_n_frames:]])
                all_temps = h5.get("bath_temperatures")
                all_temps.resize((now_n_frames,))
                all_temps[-new_n_frames:] = new_temps
                logger.info(f"Saving trajectory to existing checkpoint file {self.checkpoint_filename} ({new_n_frames} frames added; {now_n_frames} in total)")
        else:
            logger.info(f"Saving trajectory to new checkpoint file {self.checkpoint_filename} ({len(self.frames)} frames)")
            with h5py.File(self.checkpoint_filename, "w") as h5:
                h5.attrs['atomic_numbers'] = self.atomic_numbers.view(np.ndarray)
                h5.attrs['masses'] = self.masses.view(np.ndarray)
                h5.attrs['finished'] = self.finished
                h5.attrs['forwards'] = self.forwards

                n_atoms = len(self.atomic_numbers)

                energies = np.asarray([frame.energy for frame in self.frames])
                h5.create_dataset("all_energies", data=energies, maxshape=(None,),
                            compression="gzip", compression_opts=9)

                all_positions = np.stack([frame.positions for frame in self.frames])
                h5.create_dataset("all_positions", data=all_positions, maxshape=(None,n_atoms,3),
                                compression="gzip", compression_opts=9)

                all_velocities = np.stack([frame.velocities for frame in self.frames])
                h5.create_dataset("all_velocities", data=all_velocities, maxshape=(None,n_atoms,3),
                                compression="gzip", compression_opts=9)

                all_accels= np.stack([frame.accelerations for frame in self.frames])
                h5.create_dataset("all_accelerations", data=all_accels, maxshape=(None,n_atoms,3),
                                compression="gzip", compression_opts=9)

                temps = np.asarray([frame.bath_temperature for frame in self.frames])
                h5.create_dataset("bath_temperatures", data=temps, maxshape=(None,),
                            compression="gzip", compression_opts=9)

        os.remove(lockfile)

        # lower memory usage
        if keep_all:
            pass
        else:
            self.frames = [self.frames[-1]]

    def write_movie(self, filename, idxs="high"):
        if idxs == "high":
            idxs = self.high_atoms
        elif idxs == "all":
            idxs = None

        ensemble = self.as_ensemble(idxs)
        logger.info("Writing trajectory to {filename}")
        if re.search("pdb$", filename):
            cctk.PDBFile.write_ensemble_to_trajectory(filename, ensemble)
        elif re.search("mol2$", filename):
            #### connectivity matters
            for molecule in ensemble.molecules:
                molecule.assign_connectivity()
            cctk.MOL2File.write_ensemble_to_file(filename, ensemble)
        else:
            raise ValueError(f"error writing {filename}: this filetype isn't currently supported!")

    def as_ensemble(self, idxs=None):
        ensemble = cctk.ConformationalEnsemble()
        for frame in self.frames[:-1]:
            ensemble.add_molecule(frame.molecule(idxs), {"bath_temperature": frame.bath_temperature, "energy": frame.energy})
        return ensemble

    def spawn_reaction_trajectory(self, termination_function, stop_time, f_filename=None, r_filename=None):
        """
        Args:
            termination_function:
            max_time:

        Returns:
            forward ReactionTrajectory
            reverse ReactionTrajectory
        """
        if os.path.exists(f_filename) and os.path.exists(r_filename):
            logger.info(f"Trajectories to spawn already exist!")

            f_traj = ReactionTrajectory(calculator=self.calculator, integrator=self.integrator, checkpoint_filename=f_filename)
            f_traj.termination_function = termination_function
            f_traj.stop_time = stop_time

            r_traj = ReactionTrajectory(calculator=self.calculator, integrator=self.integrator, checkpoint_filename=r_filename, forwards=False)
            r_traj.termination_function = termination_function
            r_traj.stop_time = stop_time

            return f_traj, r_traj

        # add random velocity to previously inactive atoms
        frame = self.frames[frame_idx]

        random_gaussian = np.random.normal(size=frame.positions.shape).view(cctk.OneIndexedArray)
        random_gaussian[frame.active_mask()] = 0
        new_velocities = random_gaussian * np.sqrt(frame.bath_temperature * presto.constants.BOLTZMANN_CONSTANT / self.masses.reshape(-1,1))

        f_traj = ReactionTrajectory(
            timestep = self.timestep,
            atomic_numbers = self.atomic_numbers,
            high_atoms = self.high_atoms,
            active_atoms = np.array(list(range(1, len(self.atomic_numbers)+1))), # all atoms are active now
            calculator = self.calculator,
            integrator = self.integrator,
            checkpoint_filename = f_filename,
        )

        f_traj.termination_function = termination_function
        f_traj.stop_time = stop_time
        f_traj.initialize(frame=frame, new_velocities=new_velocities)

        r_traj = ReactionTrajectory(
            timestep = self.timestep,
            atomic_numbers = self.atomic_numbers,
            high_atoms = self.high_atoms,
            active_atoms = np.array(list(range(1, len(self.atomic_numbers)+1))), # all atoms are active now
            calculator = self.calculator,
            integrator = self.integrator,
            checkpoint_filename = r_filename,
        )

        r_traj.forwards = False
        r_traj.termination_function = termination_function
        r_traj.stop_time = stop_time
        r_traj.initialize(frame=frame, new_velocities=new_velocities)

        assert os.path.exists(f_filename), f"didn't save to {f_filename} successfully"
        assert os.path.exists(r_filename), f"didn't save to {r_filename} successfully"
        logger.info(f"Two new reaction trajectories spawned and saved to {f_filename} and {r_filename}")

        return f_traj, r_traj

    @classmethod
    def new_from_checkpoint(cls, checkpoint, frame):
        """
        Creates new trajectory from the given checkpoint file.

        Args:
            checkpoint (str): path to checkpoint file
            frame (int): the index of the desired frame

        Returns:
            new ``Trajectory`` object
        """
        assert isinstance(frame, int), "need an integer frame number"

        new_traj = cls(checkpoint_filename=checkpoint)
        new_traj.load_from_checkpoint(idxs=frame)

        assert len(new_traj.frames) == 1, "got too many frames!"
        return new_traj

class ReactionTrajectory(Trajectory):
    """
    Attributes:
        termination_function (function): detects if first or last Frame has reached product/SM or should otherwise be halted.
            takes ``Frame`` argument as option and returns ``True``/``False``.
            optionally, can return 1 for forward and 2 for reverse (to differentiate recrossing from productive).
        time_after_finished (float/int): how long (in fs) to continue propagation after termination conditions reached
    """

    @classmethod
    def new_from_checkpoint(self):
        pass

    def __init__(self, termination_function=None, time_after_finished=100, **kwargs):
        super().__init__(**kwargs)

        assert isinstance(time_after_finished, (int, float)), "time_after_finished must be numeric"
        self.time_after_finished = time_after_finished
        self.elapsed_since_finished = 0

        assert hasattr(termination_function, "__call__"), "termination_function must be a function!"
        self.termination_function = termination_function

    def __str__(self):
        return f"ReactionTrajectory({len(self.frames)} frames)"

    def __repr__(self):
        return f"ReactionTrajectory({len(self.frames)} frames)"

    def initialize(self, frame=None, positions=None, velocities=None, accelerations=None, bath_temp=None, new_velocities=None, **kwargs):
        """
        Generates initial frame object for reaction trajectory. Initializes any non-zero velocities.
        Velocities are taken from the Maxwell–Boltzmann distribution for the given temperature.

        Can pass either a frame object, or the relevant attributes (frame gets precedence).

        Args:
            frame (presto.frame.Frame): equilibrated frame
            new_velocities (cctk.OneIndexedArray): array of velocities to add to equilibrated frame (for previously-frozen atoms)
            velocities, accelerations, positions (cctk.OneIndexedArray): values from frame (instead of frame object)
            bath_temperature (float):

        Returns:
            frame
        """
        logger.info("Initializing new reaction trajectory...")
        if self.has_checkpoint():
            self.load_from_checkpoint()[-1]
            return

        if frame is not None:
            assert isinstance(frame, presto.frame.Frame), "need a valid frame"

            positions = frame.positions
            velocities = frame.velocities
            accelerations = frame.accelerations
            bath_temp = frame.bath_temperature

        else:
            assert positions is not None, "no Frame supplied, need positions"
            assert velocities is not None, "no Frame supplied, need velocities"
            assert accelerations is not None, "no Frame supplied, need accelerations"
            assert bath_temp is not None, "no Frame supplied, need bath temperature"

            assert isinstance(positions, cctk.OneIndexedArray)
            assert isinstance(velocities, cctk.OneIndexedArray)
            assert isinstance(accelerations, cctk.OneIndexedArray)
            assert isinstance(bath_temp, (float, int, np.integer))

        new_frame = presto.frame.Frame(self, positions, velocities, accelerations, bath_temp)

        if new_velocities is None:
            random_gaussian = np.random.normal(size=positions.shape).view(cctk.OneIndexedArray)
            random_gaussian[new_frame.active_mask()] = 0
            new_velocities = random_gaussian * np.sqrt(bath_temp * presto.constants.BOLTZMANN_CONSTANT / self.masses.reshape(-1,1))

        new_frame.velocities += new_velocities.view(cctk.OneIndexedArray)

        self.frames = [new_frame]
        self.save()

    def propagate(self, checkpoint_interval, keep_all=False, time=None):
        assert isinstance(checkpoint_interval, int) and checkpoint_interval > 0, "interval must be positive integer"

        elapsed_time = 0

        for t in np.arange(self.timestep * len(self.frames), self.stop_time, self.timestep):
            self.frames.append(self.frames[-1].next(temp=self.frames[-1].bath_temperature, forwards=self.forwards))
            if (len(self.frames) - 1) % checkpoint_interval == 0:
                self.save(keep_all=keep_all)

            elapsed_time += self.timestep
            if time is not None:
                if elapsed_time >= time:
                    return

            exit_code = self.termination_function(self.frames[-1])
            logger.debug(f"termination status:\t{t} fs\t{exit_code} code")

            if exit_code or self.elapsed_since_finished > 0:
                self.elapsed_since_finished += self.timestep
                logger.info(f"Reaction trajectory finished! {self.elapsed_since_finished:.1f} fs since termination condition met, {self.time_after_finished:.1f} fs desired.")

            if self.elapsed_since_finished >= self.time_after_finished:
                self.save(keep_all=keep_all)
                self.finished = exit_code
                return

class EquilibrationTrajectory(Trajectory):
    """
    Attributes:
        bath_scheduler (function): takes current time and returns target temperature
            alternatively, pass a number if you want a constant-temperature bath
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        bath_scheduler = kwargs.get("bath_scheduler")

        if hasattr(bath_scheduler, "__call__"):
            self.bath_scheduler = bath_scheduler
        elif isinstance(bath_scheduler, (int, float)):
            def sched(time):
                return bath_scheduler

            self.bath_scheduler = sched
        else:
            raise ValueError(f"unknown type {type(bath_scheduler)} for bath_scheduler - want either a function or a number!")

    def __str__(self):
        return f"EquilibrationTrajectory({len(self.frames)} frames)"

    def __repr__(self):
        return f"EquilibrationTrajectory({len(self.frames)} frames)"

    def initialize(self, positions, **kwargs):
        """
        Generates initial frame object for initialization trajectory.
        Velocities are taken from the Maxwell–Boltzmann distribution for the given temperature.

        Args:
            positions (cctk.OneIndexedArray):

        Returns:
            frame
        """
        logger.info("Initializing new equilibration trajectory...")
        if self.has_checkpoint():
            self.load_from_checkpoint()
            return

        assert isinstance(positions, cctk.OneIndexedArray), "positions must be a one-indexed array!"

        # move centroid to origin
        centroid = np.mean(positions, axis=0)
        positions = positions - centroid

        # determine active atoms
        inactive_mask = np.ones(shape=len(positions)).view(cctk.OneIndexedArray)
        inactive_mask[self.active_atoms] = 0
        inactive_mask  = inactive_mask.astype(bool)

        # add random velocity to everything
        sigma = np.sqrt(self.bath_scheduler(0) * presto.constants.BOLTZMANN_CONSTANT / self.masses.reshape(-1,1))
        velocities = np.random.normal(scale=sigma, size=positions.shape).view(cctk.OneIndexedArray)
        velocities[inactive_mask] = 0
#        velocities = random_gaussian / np.linalg.norm(random_gaussian, axis=1).reshape(-1,1) * sigma

        # subtract out center-of-mass translational motion
        com_translation = np.sum(self.masses.reshape(-1,1) * velocities, axis=0)
        correction_tran = np.tile(com_translation / np.sum(self.masses[self.active_atoms]), (len(velocities),1))
        correction_tran[inactive_mask] = 0
        velocities = velocities - correction_tran
        com_translation = np.sum(self.masses.reshape(-1,1) * velocities, axis=0)
        assert np.linalg.norm(np.sum(self.masses.reshape(-1,1) * velocities, axis=0)) < 0.0001, "didn't remove COM translation well enough!"

        velocities = velocities.view(cctk.OneIndexedArray)
        accelerations = np.zeros_like(positions).view(cctk.OneIndexedArray)
        self.frames = [presto.frame.Frame(self, positions, velocities, accelerations, self.bath_scheduler(0))]
        self.save()

    def propagate(self, checkpoint_interval, keep_all=False, time=None):
        assert isinstance(checkpoint_interval, int) and checkpoint_interval > 0, "interval must be positive integer"

        elapsed_time = 0
        for t in np.arange(self.timestep * self.num_frames(), self.stop_time, self.timestep):

            self.frames.append(self.frames[-1].next(temp=self.bath_scheduler(t)))
            if (len(self.frames) - 1) % checkpoint_interval == 0:
                self.save(keep_all=keep_all)

            elapsed_time += self.timestep
            if time is not None:
                if elapsed_time >= time:
                    return
        self.save(keep_all=keep_all)
        self.finished = True


def join(traj1, traj2):
    """
    Join two reaction trajectories together -- one forward and one reverse!
    Returns a single reaction trajectory.

    Args:
        traj1 (presto.ReactionTrajectory): forward trajectory
        traj2 (presto.ReactionTrajectory): reverse trajectory

    Returns:
        combined ``ReactionTrajectory``
    """
    logger.info("Joining forward and reverse reaction trajectories....")
    assert isinstance(traj1, ReactionTrajectory), "Need a ReactionTrajectory"
    assert isinstance(traj2, ReactionTrajectory), "Need a ReactionTrajectory"

    assert traj1.forwards == True, "First trajectory must be forwards!"
    assert traj2.forwards == False, "Second trajectory must be reverse!"

    assert traj1.finished, "First trajectory must be finished!"
    assert traj2.finished, "Second trajectory must be finished!"

    assert np.array_equal(traj1.frames[0].positions, traj2.frames[0].positions), "Link positions must be same!"
    assert np.array_equal(traj1.frames[0].velocities, traj2.frames[0].velocities), "Link velocities must be same!"
    assert np.array_equal(traj1.frames[0].accelerations, traj2.frames[0].accelerations), "Link accelerations must be same!"

    new_traj = ReactionTrajectory(
        timestep = traj1.timestep,
        atomic_numbers = traj1.atomic_numbers,
        high_atoms = traj1.high_atoms,
        active_atoms = traj1.active_atoms,
        calculator = traj1.calculator,
        integrator = traj1.integrator,
        termination_function = traj1.termination_function,
        stop_time = traj1.stop_time,
        forwards = True,
    )

    f_frames = copy.deepcopy(traj1.frames)
    r_frames = copy.deepcopy(traj2.frames)
    r_frames.reverse()

    new_traj.frames = r_frames + f_frames

    if traj1.finished == 2:
        #### if the first traj finished with the reverse condition, reverse order
        new_traj.frames = new_traj.frames[::-1]

    return new_traj





import numpy as np
import math, copy, cctk
from scipy import constants

import presto

class Integrator():
    def next(self, frame, forwards=True):
        pass

class VelocityVerletIntegrator(Integrator):
    def next(self, frame, forwards=True):
        calculator = frame.trajectory.calculator
        timestep = frame.trajectory.timestep
        if forwards == False:
            timestep = timestep * -1

        x_full = frame.positions + frame.velocities * timestep + 0.5 * frame.accelerations * (timestep ** 2)

        energy, forces = calculator.evaluate(frame.trajectory.atomic_numbers, x_full, frame.trajectory.high_atoms)
        forces[frame.inactive_mask()] = 0

        a_full = forces / frame.masses()

        v_full = frame.velocities + (frame.accelerations + a_full) * 0.5 * timestep

        return energy, x_full, v_full, a_full


class LangevinIntegrator(VelocityVerletIntegrator):
    """
    Defines a Langevin integrator (NVT ensemble).

    Attributes:
        viscosity (float): solvent viscosity (in amu/(fs*Å))
        radius (float): distance from center where this kicks in!
        maxiter (int): maximum number of numerical drag iterations
        tolerance (float): convergence cutoff for numerical drag calculation - smaller is tighter
        potential (float): additional function that maps positions to forces
            needs to return ``cctk.OneIndexedArray``
    """

    def __init__(self, viscosity=0.001, convert_from_pascal_seconds=True, radius=0, potential=None):
        assert isinstance(viscosity, (int, float)), "viscosity must be numeric"
        if convert_from_pascal_seconds:
            viscosity = viscosity * presto.constants.AMU_A_FS_PER_PASCAL_SECOND
        self.viscosity = viscosity

        assert isinstance(radius, (int, float)), "radius must be numeric"
        self.radius = radius

        # if you use this, it's your own problem to check that you did it right!
        self.potential = potential

    def old_next(self, frame, forwards=True):
        calculator = frame.trajectory.calculator
        timestep = frame.trajectory.timestep
        if forwards == False:
            timestep = timestep * -1

        v_half = frame.velocities + 0.5 * frame.accelerations * timestep

        x_full = frame.positions + v_half * timestep

        energy, forces = calculator.evaluate(frame.trajectory.atomic_numbers, x_full, frame.trajectory.high_atoms)

        #### compute random forces
        xi = 6 * math.pi * self.viscosity * frame.radii()
        variance =  2 * xi * presto.constants.BOLTZMANN_CONSTANT * frame.bath_temperature / frame.trajectory.timestep
        no_apply_to = np.linalg.norm(x_full, axis=1) < self.radius

        random_forces = np.random.normal(loc=0, scale=np.sqrt(variance.reshape(-1,1)), size=x_full.shape)
        random_forces[no_apply_to,:] = 0
        forces += random_forces.view(cctk.OneIndexedArray)

        if self.potential is not None:
            forces += self.potential(x_full)

        #### compute drag
        drag = np.zeros_like(random_forces).view(cctk.OneIndexedArray)
        v_full = v_half + 0.5 * (forces / frame.masses()) * timestep

        #### have to do numerical convergence since drag is proportional to instantaneous velocity
        for n in range(self.maxiter):
            prev_v = v_full
            drag = -1 * xi.reshape(-1,1) * v_full
            v_full = v_half + 0.5 * ((forces + drag) / frame.masses()) * timestep

            #### if we've converged, return - otherwise iterate again
            if np.mean(np.linalg.norm(prev_v - v_full, axis=0)/np.linalg.norm(prev_v, axis=0)) < self.tolerance:
                break

        drag.view(np.ndarray)[no_apply_to,:] = 0
        forces += drag

        forces[frame.inactive_mask()] = 0
        a_full = forces / frame.masses()

        return energy, x_full, v_full, a_full

    def next(self, frame, forwards=True):
        """
        Using the approach from http://itf.fys.kuleuven.be/~enrico/Teaching/molecular_dynamics_2015.pdf
        """
        calculator = frame.trajectory.calculator
        timestep = frame.trajectory.timestep
        if forwards == False:
            timestep = timestep * -1

        xi = 6 * math.pi * self.viscosity * frame.radii() / frame.trajectory.masses
        sigma = np.sqrt(2 * xi * presto.constants.BOLTZMANN_CONSTANT * frame.bath_temperature / frame.trajectory.masses)

        # exclude those beyond the radius
        no_apply_to = np.linalg.norm(frame.positions, axis=1) < self.radius
        xi[no_apply_to] = 0
        sigma[no_apply_to] = 0
        xi = xi.reshape(-1,1)
        sigma = sigma.reshape(-1,1)

        rand1 = np.random.normal(size=frame.positions.shape).view(cctk.OneIndexedArray)
        rand2 = np.random.normal(size=frame.positions.shape).view(cctk.OneIndexedArray)

        C = 0.5 * (timestep ** 2) * (frame.accelerations - (xi * frame.velocities)) + sigma * (timestep ** 1.5) * ((0.5 * rand1) + (rand2 / (2 * math.sqrt(3))))

        x_full = frame.positions + timestep * frame.velocities + C

        energy, forces = calculator.evaluate(frame.trajectory.atomic_numbers, x_full, frame.trajectory.high_atoms)
        forces[frame.inactive_mask()] = 0
        a_full = forces / frame.masses()

        v_full = frame.velocities + 0.5 * timestep * (a_full + frame.accelerations) - timestep * xi * frame.velocities + sigma * math.sqrt(timestep) * rand1 - xi * C

        return energy, x_full, v_full, a_full


def spherical_harmonic_potential(radius, force_constant=0.004184):
    """
    Default force constant is 10 kcal/mol per Å**2, akin to that employed by Singleton (JACS, 2016, 138, 15167).

    In *presto* units, this is 0.004184 amu Å**2 fs**-2.

    Args:
        radius (float): area outside which this takes effect
        force_constant (float):

    Returns:
        function mapping positions to forces with the desired force constant and radius
    """
    assert isinstance(radius, (int, float))
    assert isinstance(force_constant, (int, float))
    assert radius > 0
    assert force_constant > 0

    def force(positions):
        radii = np.linalg.norm(positions, axis=1)
        forces = -0.5 * force_constant * (positions - positions/radii.reshape(-1,1) * radius)
        forces = forces * np.linalg.norm(positions - positions/radii.reshape(-1,1) * radius, axis=1).reshape(-1,1)
        inside = radii < radius
        forces.view(np.ndarray)[inside,:] = 0
        return forces.view(cctk.OneIndexedArray)

    return force



import numpy as np
import math, copy, cctk

import cctk
from cctk.helper_functions import get_symbol, get_vdw_radius

import presto

class Frame():
    """
    Represents one frame in a trajectory.

    Attributes:
        trajectory (presto.Trajectory):
        positions (cctk.OneIndexedArray):
        velocities (cctk.OneIndexedArray):
        accelerations (cctk.OneIndexedArray):

        energy (float):
        bath_temperature (float):
    """

    def __init__(self, trajectory, x, v, a, bath_temperature=298, energy=0.0):
        assert isinstance(trajectory, presto.trajectory.Trajectory), "need trajectory"

        assert len(x) == len(v), "length of positions not same as length of velocities!"
        assert len(x) == len(a), "length of positions not same as length of accelerations!"

        assert isinstance(x, cctk.OneIndexedArray), "positions is not a one-indexed array!"
        assert isinstance(v, cctk.OneIndexedArray), "velocities is not a one-indexed array!"
        assert isinstance(a, cctk.OneIndexedArray), "accelerations is not a one-indexed array!"

        assert (x.ndim == 2) and (x.shape[1] ==  3), "positions must be an n x 3 ndarray"
        assert (v.ndim == 2) and (v.shape[1] ==  3), "velocities must be an n x 3 ndarray"
        assert (a.ndim == 2) and (a.shape[1] ==  3), "accelerations must be an n x 3 ndarray"

        assert isinstance(bath_temperature, (float, int, np.integer)), "bath temperature needs to be numeric!"
        assert bath_temperature >= 0, "bath temperature must be positive or 0"

        self.trajectory = trajectory
        self.positions = x
        self.velocities = v
        self.accelerations = a
        self.bath_temperature = bath_temperature
        self.energy = energy

    def __str__(self):
        temp = f"E={self.energy}, temp={self.bath_temperature}\n"
        n_atoms = len(self.positions)
        for atom in range(1,n_atoms+1):
            x,v,a = self.positions[atom],self.velocities[atom],self.accelerations[atom]
            temp += f"{atom:3d} [ {x[1]:8.3f} {x[2]:8.3f} {x[3]:8.3f} ] [ {v[1]:8.3f} {v[2]:8.3f} {v[3]:8.3f} ] [ {a[1]:10.2E} {a[2]:10.2E} {a[3]:10.2E} ]\n"
        return temp[:-1]

    def next(self, temp=None, forwards=True):
        """
        Computes next frame using ``self.trajectory.integrator``.

        The desired bath temperature is not used in the current force calculations, but is passed to the output frame.
        """
        if temp is None:
            temp = self.bath_temperature
        assert isinstance(temp, (float, int, np.integer)), "temp must be numeric!"

        integrator = self.trajectory.integrator
        energy, new_x, new_v, new_a = integrator.next(self, forwards=forwards)
        self.energy = energy
        return Frame(self.trajectory, new_x, new_v, new_a, temp)

    def prev(self, temp=None):
        """
        Computes previous frame using ``self.trajectory.integrator``.

        The desired bath temperature is not used in the current force calculations, but is passed to the output frame.
        """
        return self.next(temp, forwards=False)

    def temperature(self):
        """
        Computes the instantaneous temperature based on the equipartition theorem, counting only the active atoms.

        T = sum{ m_i * v_i ** 2} / (kB * Nf)
        """
        v = np.linalg.norm(self.velocities[self.trajectory.active_atoms], axis=1)
        m = self.trajectory.masses.view(cctk.OneIndexedArray)[self.trajectory.active_atoms]
        K = np.multiply(m.view(np.ndarray), np.power(v, 2))
        F = 3 * len(self.trajectory.active_atoms) - 3

        if F == 0: # single particle
            F = 3

        return float(np.sum(K) / (F * presto.constants.BOLTZMANN_CONSTANT))

    def pressure(self):
        """
        Computes the pressure based on the following formula:

        P = 1/(3*V) * (\sum{m_i * v_i * v_i + r_i * f_i}
        """
        m = self.trajectory.masses.view(cctk.OneIndexedArray)[self.trajectory.active_atoms].view(np.ndarray)
        v = self.velocities[self.trajectory.active_atoms].view(np.ndarray)
        a = self.accelerations[self.trajectory.active_atoms].view(np.ndarray)
        x = self.positions[self.trajectory.active_atoms].view(np.ndarray)

        volume = self.volume()

        v = np.linalg.norm(self.velocities[self.trajectory.active_atoms], axis=1)
        m = self.trajectory.masses.view(cctk.OneIndexedArray)[self.trajectory.active_atoms]
        K = np.multiply(m.view(np.ndarray), np.power(v, 2))

        pressure = np.sum(K) / (3 * volume)
        for pos, mass, accel in zip(x, m, a):
            pressure += np.dot(pos, mass * accel) / (3 * volume)

        return pressure / presto.constants.AMU_A_FS2_PER_ATM

    def volume(self):
        return self.molecule().volume()

    def inactive_mask(self):
        """
        Returns an ``np.ndarray`` of the same length as ``positions`` where every active atom is ``False`` and every inactive atom is ``True``.
        """
        inactive_mask = np.ones(shape=len(self.positions)).view(cctk.OneIndexedArray)
        inactive_mask[self.trajectory.active_atoms] = 0
        inactive_mask  = inactive_mask.astype(bool)
        return inactive_mask

    def active_mask(self):
        """
        Returns an ``np.ndarray`` of the same length as ``positions`` where every active atom is ``False`` and every inactive atom is ``True``.
        """
        active_mask = np.zeros(shape=len(self.positions)).view(cctk.OneIndexedArray)
        active_mask[self.trajectory.active_atoms] = 1
        active_mask  = active_mask.astype(bool)
        return active_mask

    def molecule(self, idxs=None):
        if idxs is not None:
            return cctk.Molecule(self.trajectory.atomic_numbers[idxs], self.positions[idxs])
        else:
            return cctk.Molecule(self.trajectory.atomic_numbers, self.positions)

    def remove_com_motion(self):
        # move centroid to origin
        centroid = np.mean(self.positions, axis=0)
        self.positions = self.positions - centroid
        assert np.linalg.norm(np.sum(self.positions, axis=0)) < 0.0001, "didn't center well enough!"

        #### subtract out center-of-mass translational motion (linear momentum)
        com_translation = np.sum(self.masses() * self.velocities, axis=0)
        correction_tran = np.tile(com_translation / np.sum(self.masses()), (len(self.velocities),1))
        self.velocities = self.velocities - correction_tran
        assert np.linalg.norm(np.sum(self.masses() * self.velocities, axis=0)) < 0.0001, "didn't remove COM translation well enough!"

        return self

        #### subtract out center-of-mass rotational motion - not working
        com_rotation = np.sum(np.cross(self.velocities, self.masses() * self.positions), axis=0)
        print(com_rotation)

        repeat_com_rotation = np.tile(com_rotation, (self.positions.shape[0], 1))
        parallel_position_vectors = np.dot(self.positions, com_rotation).reshape(-1,1) / np.dot(com_rotation, com_rotation) * repeat_com_rotation
        perp_position_vectors = self.positions - parallel_position_vectors

        correction = np.cross(
            perp_position_vectors,
            com_rotation,
        ).view(cctk.OneIndexedArray)

        scale = np.power(np.linalg.norm(perp_position_vectors, axis=1), 2).reshape(-1,1) * self.masses()
        scale = self.masses()
        correction = correction / scale

        correction_applied = np.sum(np.cross(correction, self.masses() * self.positions), axis=0)
        correction = correction * np.linalg.norm(com_rotation) / (np.linalg.norm(correction_applied))
        self.velocities = self.velocities - correction.view(cctk.OneIndexedArray)

        assert np.linalg.norm(np.sum(self.masses() * self.velocities, axis=0)) < 0.0001, "COM rotation correction introduced translation!"
        assert np.linalg.norm(np.sum(np.cross(self.velocities, self.masses() * self.positions), axis=0)) < 0.0001, "didn't remove COM rotation well enough!"
        return self

    def masses(self):
        return self.trajectory.masses.reshape(-1,1).view(cctk.OneIndexedArray)

    def radii(self):
        vdw_radii = {z: get_vdw_radius(z) for z in set(self.trajectory.atomic_numbers)}
        radii_by_num = np.array([vdw_radii[z] for z in self.trajectory.atomic_numbers]).view(cctk.OneIndexedArray)
        return radii_by_num

    def L(self):
        """ Angular momentum """
        return np.sum(np.cross(self.velocities, self.masses() * self.positions), axis=0)
